@@ -19,7 +19,6 @@
 
 from functools import total_ordering
 
-import lazy_object_proxy
 import marshmallow.fields as fields
 from marshmallow.base import SchemaABC
 from marshmallow import class_registry, utils
@@ -27,17 +26,20 @@ from marshmallow.exceptions import ValidationError
 import logging
 import copy
 
+import rdflib
+
 import typing
 import types
 
-from calamus.utils import normalize_type, normalize_value, Proxy
+from calamus.schema import JsonLDSchema
+from calamus.utils import normalize_type, normalize_value, ONTOLOGY_QUERY, Proxy
 
 logger = logging.getLogger("calamus")
 
 
 @total_ordering
 class IRIReference(object):
-    """ Represent an IRI in a namespace.
+    """Represent an IRI in a namespace.
 
     Args:
         namespace (Namespace): The ``Namespace`` this IRI is part of.
@@ -77,13 +79,27 @@ class Namespace(object):
 
     Args:
         namespace (str): The base namespace URI for this namespace.
+        ontology (str): Path to an ontology(OWL) file for this namespace.
     """
 
-    def __init__(self, namespace):
+    def __init__(self, namespace, ontology=None):
         self.namespace = namespace
+        self.ontology = None
+
+        if ontology:
+            g = rdflib.Graph()
+            self.ontology = g.parse(ontology)
 
     def __getattr__(self, name):
-        return IRIReference(self, name)
+        reference = IRIReference(self, name)
+
+        if self.ontology:
+            p = rdflib.URIRef(str(reference))
+            qres = self.ontology.query(ONTOLOGY_QUERY, initBindings={"property": p})
+            if not next(iter(qres), False):
+                raise ValueError(f"Property {name} does not exist in namespace {self.namespace}")
+
+        return reference
 
     def __str__(self):
         return self.namespace
@@ -106,6 +122,7 @@ class _JsonLDField(fields.Field):
         self.reverse = kwargs.get("reverse", False)
         self.init_name = kwargs.get("init_name", None)
         self.add_value_types = kwargs.get("add_value_types", False)
+        self.default = kwargs.get("default", None)
 
     @property
     def data_key(self):
@@ -181,7 +198,8 @@ class IRI(String):
 
     def _serialize(self, value, attr, obj, **kwargs):
         value = super()._serialize(value, attr, obj, **kwargs)
-        return {"@id": value}
+        if value:
+            return {"@id": value}
 
     def _deserialize(self, value, attr, data, **kwargs):
         if "@id" in value:
@@ -261,6 +279,58 @@ class DateTime(_JsonLDField, fields.DateTime):
         raise self.make_error("invalid", input=value, obj_type=self.OBJ_TYPE)
 
 
+class NaiveDateTime(_JsonLDField, fields.NaiveDateTime):
+    """A naive date/time field."""
+
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+
+    def _serialize(self, value, attr, obj, **kwargs):
+        value = super()._serialize(value, attr, obj, **kwargs)
+        if self.parent.opts.add_value_types or self.add_value_types:
+            value = {"@value": value, "@type": "http://www.w3.org/2001/XMLSchema#dateTime"}
+        return value
+
+
+class AwareDateTime(_JsonLDField, fields.AwareDateTime):
+    """A naive date/time field."""
+
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+
+    def _serialize(self, value, attr, obj, **kwargs):
+        value = super()._serialize(value, attr, obj, **kwargs)
+        if self.parent.opts.add_value_types or self.add_value_types:
+            value = {"@value": value, "@type": "http://www.w3.org/2001/XMLSchema#dateTime"}
+        return value
+
+
+class Time(_JsonLDField, fields.Time):
+    """A naive date/time field."""
+
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+
+    def _serialize(self, value, attr, obj, **kwargs):
+        value = super()._serialize(value, attr, obj, **kwargs)
+        if self.parent.opts.add_value_types or self.add_value_types:
+            value = {"@value": value, "@type": "http://www.w3.org/2001/XMLSchema#time"}
+        return value
+
+
+class Date(_JsonLDField, fields.Date):
+    """A naive date/time field."""
+
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+
+    def _serialize(self, value, attr, obj, **kwargs):
+        value = super()._serialize(value, attr, obj, **kwargs)
+        if self.parent.opts.add_value_types or self.add_value_types:
+            value = {"@value": value, "@type": "http://www.w3.org/2001/XMLSchema#date"}
+        return value
+
+
 class Dict(_JsonLDField, fields.Dict):
     """A dict field."""
 
@@ -273,6 +343,54 @@ class Raw(_JsonLDField, fields.Raw):
     pass
 
 
+class RawJsonLD(_JsonLDField, fields.Raw):
+    """A raw JSON-LD field."""
+
+    def _dereference_single_id(self, value, attr, **kwargs):
+        """Dereference a single id."""
+        data = kwargs["_all_objects"].get(value, None)
+        if not data:
+            raise ValueError("Couldn't dereference id {id}".format(id=value))
+
+        try:
+            data = dict(data)
+        except (TypeError, ValueError):
+            raise ValueError(f"Couldn't convert value '{data}' to dictionary.")
+
+        if self.reverse:
+            # we need to remove the property from the child when handling reverse nesting
+            del data[attr]
+
+        return data
+
+    def _dereference_flattened(self, value, attr, **kwargs):
+        """Dereference an id or a list of ids."""
+        if isinstance(value, list) or isinstance(value, types.GeneratorType):
+            return [self._dereference_flattened(i, attr, **kwargs) for i in value]
+        if isinstance(value, str):
+            return self._dereference_single_id(value, attr, **kwargs)
+        elif isinstance(value, dict):
+            if len(value.keys()) == 1 and "@id" in value:
+                value = self._dereference_single_id(value["@id"], attr, **kwargs)
+                for k, v in value.items():
+                    if not k.startswith("@"):
+                        value[k] = self._dereference_flattened(v, attr, **kwargs)
+                return value
+            else:
+                return value
+        else:
+            raise ValueError("Nested field needs to be a dict or an id entry/list, got {value}".format(value=value))
+
+    def _deserialize(self, value, attr, data, **kwargs):
+        """Deserialize object."""
+
+        if kwargs.get("flattened", False):
+            # could be id references, dereference them to continue deserialization
+            value = self._dereference_flattened(value, attr, **kwargs)
+
+        return super()._deserialize(value, attr, data, **kwargs)
+
+
 class Nested(_JsonLDField, fields.Nested):
     """A reference to one or more nested classes."""
 
@@ -282,6 +400,19 @@ class Nested(_JsonLDField, fields.Nested):
         if not isinstance(self.nested, list):
             self.nested = [self.nested]
 
+        nested = []
+
+        for n in self.nested:
+            if isinstance(n, str) or issubclass(n, JsonLDSchema):
+                nested.append(n)
+            else:
+                if hasattr(n, "__calamus_schema__"):
+                    nested.append(n.__calamus_schema__)
+                else:
+                    raise ValueError(f"Only calamus schema is allowed in nested fields, not {n}")
+
+        self.nested = nested
+
     @property
     def schema(self):
         """The nested Schema object.
@@ -319,7 +450,7 @@ class Nested(_JsonLDField, fields.Nested):
                     if isinstance(nest, type) and issubclass(nest, SchemaABC):
                         schema_class = nest
                     elif not isinstance(nest, (str, bytes)):
-                        raise ValueError("Nested fields must be passed a " "Schema, not {}.".format(nest.__class__))
+                        raise ValueError("Nested fields must be passed a Schema, not {}.".format(nest.__class__))
                     elif nest == "self":
                         ret = self
                         while not isinstance(ret, SchemaABC):
@@ -413,12 +544,12 @@ class Nested(_JsonLDField, fields.Nested):
 
         if not schema:
             ValueError("Type {} not found in {}.{}".format(value["@type"], type(self.parent), self.data_key))
-
-        if schema.lazy:
-            return Proxy(lambda: schema.load(value, unknown=self.unknown, partial=partial), schema, value)
         if not schema._all_objects and self.root._all_objects:
             schema._all_objects = self.root._all_objects
         schema._reversed_properties = self.root._reversed_properties
+
+        if schema.lazy:
+            return Proxy(lambda: schema.load(value, unknown=self.unknown, partial=partial), schema, value)
         return schema.load(value, unknown=self.unknown, partial=partial)
 
     def _load(self, value, data, partial=None, many=False):
