@@ -22,15 +22,16 @@ from marshmallow.utils import missing, is_collection, RAISE, set_value, EXCLUDE,
 from marshmallow import post_load
 from collections.abc import Mapping
 from marshmallow.error_store import ErrorStore
+from rdflib.plugins.sparql import prepareQuery
 from uuid import uuid4
-
+import rdflib
 from pyld import jsonld
 
 import inspect
-
 import typing
+from functools import lru_cache
 
-from calamus.utils import normalize_id, normalize_type, Proxy
+from calamus.utils import normalize_id, normalize_type, Proxy, validate_field_properties
 
 _T = typing.TypeVar("_T")
 
@@ -214,6 +215,9 @@ class JsonLDSchema(Schema, metaclass=JsonLDSchemaMeta):
         """
         ret = []
 
+        if not self._all_objects:
+            return ret
+
         for d in self._all_objects.values():
             if field_name not in d:
                 continue
@@ -233,7 +237,10 @@ class JsonLDSchema(Schema, metaclass=JsonLDSchemaMeta):
 
     def _deserialize(
         self,
-        data: typing.Union[typing.Mapping[str, typing.Any], typing.Iterable[typing.Mapping[str, typing.Any]],],
+        data: typing.Union[
+            typing.Mapping[str, typing.Any],
+            typing.Iterable[typing.Mapping[str, typing.Any]],
+        ],
         *args,
         error_store: ErrorStore,
         many: bool = False,
@@ -251,7 +258,7 @@ class JsonLDSchema(Schema, metaclass=JsonLDSchemaMeta):
             for d in data:
                 self._all_objects[d["@id"]] = d
 
-                if self._compare_ids(d["@type"], self.opts.rdf_type):
+                if "@type" in d and self._compare_ids(d["@type"], self.opts.rdf_type):
                     new_data.append(d)
 
             data = new_data
@@ -328,7 +335,11 @@ class JsonLDSchema(Schema, metaclass=JsonLDSchemaMeta):
                 d_kwargs["lazy"] = self.lazy
                 getter = lambda val: field_obj.deserialize(val, field_name, data, **d_kwargs)
                 value = self._call_and_store(
-                    getter_func=getter, data=raw_value, field_name=field_name, error_store=error_store, index=index,
+                    getter_func=getter,
+                    data=raw_value,
+                    field_name=field_name,
+                    error_store=error_store,
+                    index=index,
                 )
                 if value is not missing:
                     key = field_obj.attribute or attr_name
@@ -357,7 +368,9 @@ class JsonLDSchema(Schema, metaclass=JsonLDSchemaMeta):
                         set_value(typing.cast(typing.Dict, ret), key, value)
                     elif unknown == RAISE:
                         error_store.store_error(
-                            [self.error_messages["unknown"]], key, (index if index_errors else None),
+                            [self.error_messages["unknown"]],
+                            key,
+                            (index if index_errors else None),
                         )
 
         self._init_names_mapping = {
@@ -366,6 +379,73 @@ class JsonLDSchema(Schema, metaclass=JsonLDSchemaMeta):
 
         return ret
 
+    def validate_properties(self, data, ontology, return_valid_data=False, strict=False):
+        """Validate JSON-LD against an ontology.
+
+        Args:
+            data (Union[object, dict, list]): JSON-LD data or model (or list of them).
+            ontology (str): Path/URI to an ontology file.
+            return_valid_data (bool): Whether to delete invalid properties to return only valid data or else
+                returns a dict containing valid and invalid properties, Default: ``False``
+        """
+
+        if isinstance(data, self.Meta.model) or all(isinstance(s, self.Meta.model) for s in data):
+            data = self.dump(data)
+
+        g = rdflib.Graph()
+        if not isinstance(ontology, list):
+            ontology = [ontology]
+
+        for o in ontology:
+            g.parse(o)
+
+        # NOTE: the query checks if the property we are passing is a property defined in the ontology
+        q = prepareQuery(
+            "ask { { ?property rdf:type <http://www.w3.org/2002/07/owl#DatatypeProperty> .} UNION { ?property rdf:type "
+            "<http://www.w3.org/2002/07/owl#ObjectProperty> .} }"
+        )
+
+        if self.many:
+            i = 0
+            # NOTE: res helps with memoization and is also the return value if return_valid_data is False
+            res = {"valid": set(), "invalid": set()}
+
+            valdata = []
+            for obj in data:
+                fres = validate_field_properties(obj, g, query=q, mem=res)
+                res["valid"] = res["valid"].union(fres["valid"])
+                res["invalid"] = res["invalid"].union(fres["invalid"])
+
+                if return_valid_data:
+                    resf = obj.copy()
+                    valdata.append(resf)
+                    for inval in fres["invalid"]:
+                        valdata[i].pop(inval, None)
+                    i += 1
+
+            if strict and res["invalid"]:
+                invalid_props = ", ".join(res["invalid"])
+                raise ValueError(f"Invalid properties found in ontology: {invalid_props}")
+
+            if return_valid_data:
+                return valdata
+
+            return res
+
+        res = validate_field_properties(data, g, query=q)
+
+        if strict and res["invalid"]:
+            invalid_props = ", ".join(res["invalid"])
+            raise ValueError(f"Invalid properties found in ontology: {invalid_props}")
+
+        if return_valid_data:
+            resd = data.copy()
+            for inv in res["invalid"]:
+                resd.pop(inv, None)
+            return resd
+
+        return res
+
     def _reversed_fields(self):
         """Get fields that are reversed in type hierarchy."""
         if hasattr(self, "_reversed_properties"):
@@ -396,13 +476,17 @@ class JsonLDSchema(Schema, metaclass=JsonLDSchemaMeta):
         kwargs = {}
         has_kwargs = False
         for _, parameter in const_args.parameters.items():
-            if parameter.kind in [inspect.Parameter.POSITIONAL_OR_KEYWORD, inspect.Parameter.POSITIONAL_ONLY]:
+            if parameter.kind is inspect.Parameter.POSITIONAL_ONLY:
+                # NOTE: To avoid potential errors we require positional-only arguments to always be present in data.
                 if parameter.name not in keys:
                     raise ValueError("Field {} not found in data {}".format(parameter.name, data))
                 args.append(data[parameter.name])
                 keys.remove(parameter.name)
-            elif parameter.kind is inspect.Parameter.KEYWORD_ONLY:
-                if parameter.name in keys:
+            elif parameter.kind in [inspect.Parameter.POSITIONAL_OR_KEYWORD, inspect.Parameter.KEYWORD_ONLY]:
+                if parameter.name not in keys:
+                    if parameter.default is inspect.Parameter.empty:
+                        raise ValueError("Field {} not found in data {}".format(parameter.name, data))
+                else:
                     kwargs[parameter.name] = data[parameter.name]
                     keys.remove(parameter.name)
             elif parameter.kind is inspect.Parameter.VAR_KEYWORD:
@@ -429,3 +513,78 @@ class JsonLDSchema(Schema, metaclass=JsonLDSchemaMeta):
             )
 
         return instance
+
+
+class JsonLDAnnotation(type):
+    """Meta-class allowing automated generation of calamus schema based on annotations.
+
+    Example:
+
+    .. code-block:: python
+
+       import datetime.datetime as dt
+
+       from calamus import JsonLDAnnotation
+       import calamus.fields as fields
+
+       schema = fields.Namespace("http://schema.org/")
+
+       class User(metaclass=JsonLDAnnotation):
+           class Meta:
+               rdf_type = schema.Person
+           _id = fields.Id()
+           birth_date = fields.Date(schema.birthDate, default=dt.now)
+           name = fields.String(schema.name, default=lambda: "John")
+
+        user = User()
+
+        # dumping
+        User.schema().dump(user)
+        # or
+        user.dump()
+
+        # loading
+        u = User.schema().load({"_id": "http://example.com/user/1", "name": "Bill", "birth_date": "1970-01-01 00:00"})
+    """
+
+    def __new__(mcs, name, bases, namespace, **kwargs):
+        import calamus.fields as fields
+
+        attribute_dict = {}
+        for attr_name, value in namespace.copy().items():
+            if isinstance(value, fields._JsonLDField):
+                attribute_dict[attr_name] = value
+
+                if hasattr(value, "default"):
+                    if callable(value.default):
+                        namespace[attr_name] = value.default()
+                    else:
+                        namespace[attr_name] = value.default
+                else:
+                    del namespace[attr_name]
+
+        if "Meta" not in namespace or not hasattr(namespace["Meta"], "rdf_type"):
+            raise ValueError("Setting 'rdf_type' on the `class Meta` is required for calamus annotations")
+
+        attribute_dict["Meta"] = type("Meta", (), {"rdf_type": namespace["Meta"].rdf_type})
+        namespace["__calamus_schema__"] = type(f"{name}Schema", (JsonLDSchema,), attribute_dict)
+
+        @lru_cache(maxsize=5)
+        def schema(*args, **kwargs):
+            """Convenience method to access calamus schema of a class."""
+            return namespace["__calamus_schema__"](*args, **kwargs)
+
+        namespace[schema.__name__] = schema
+
+        def dump(self, *args, **kwargs):
+            """Convenience method to dump object directly."""
+            return type(self).schema(*args, **kwargs).dump(self)
+
+        namespace[dump.__name__] = dump
+
+        cls = super().__new__(mcs, name, bases, namespace, **kwargs)
+
+        namespace["__calamus_schema__"].Meta.model = cls
+        namespace["__calamus_schema__"].opts.model = cls
+
+        return cls
